@@ -1,6 +1,7 @@
 /* common_func.c - functions used almost everywhere */
 
 #include "common_func.h"
+#include "output.h"
 #include "parse_cmdline.h"
 #include "version.h"
 #include "win_utils.h"
@@ -331,32 +332,6 @@ void rsh_exit(int code)
 	exit(code);
 }
 
-/*=========================================================================
- * Error reporting functions
- *=========================================================================*/
-
-static void report_error_default(const char* srcfile, int srcline,
-	const char* format, ...);
-
-void (*rsh_report_error)(const char* srcfile, int srcline,
-	const char* format, ...) = report_error_default;
-
-/**
- * Print given library failure to stderr.
- *
- * @param srcfile source file to report error on fail
- * @param srcline source code line to be reported on fail
- * @param format printf-formatted error message
- */
-static void report_error_default(const char* srcfile, int srcline, const char* format, ...)
-{
-	va_list ap;
-	rsh_fprintf(stderr, "RHash: error at %s:%u: ", srcfile, srcline);
-	va_start(ap, format);
-	rsh_vfprintf(stderr, format, ap); /* report the error to stderr */
-	va_end(ap);
-}
-
 /*=========================================================================
  * Memory functions
  *=========================================================================*/
@@ -372,10 +347,7 @@ static void report_error_default(const char* srcfile, int srcline, const char* f
 void* rhash_malloc(size_t size, const char* srcfile, int srcline)
 {
 	void* res = malloc(size);
-	if (!res) {
-		rsh_report_error(srcfile, srcline, "%s(%u) failed\n", "malloc", (unsigned)size);
-		rsh_exit(2);
-	}
+	RSH_REQUIRE_IMPL(res, srcfile, srcline, "%malloc(%u) failed\n", (unsigned)size);
 	return res;
 }
 
@@ -391,10 +363,7 @@ void* rhash_malloc(size_t size, const char* srcfile, int srcline)
 void* rhash_calloc(size_t num, size_t size, const char* srcfile, int srcline)
 {
 	void* res = calloc(num, size);
-	if (!res) {
-		rsh_report_error(srcfile, srcline, "calloc(%u, %u) failed\n", (unsigned)num, (unsigned)size);
-		rsh_exit(2);
-	}
+	RSH_REQUIRE_IMPL(res, srcfile, srcline, "calloc(%u, %u) failed\n", (unsigned)num, (unsigned)size);
 	return res;
 }
 
@@ -415,11 +384,7 @@ char* rhash_strdup(const char* str, const char* srcfile, int srcline)
 	char* res = (char*)malloc(strlen(str)+1);
 	if (res) strcpy(res, str);
 #endif
-
-	if (!res) {
-		rsh_report_error(srcfile, srcline, "strdup(\"%s\") failed\n", str);
-		rsh_exit(2);
-	}
+	RSH_REQUIRE_IMPL(res, srcfile, srcline, "strdup(\"%s\") failed\n", str);
 	return res;
 }
 
@@ -440,11 +405,7 @@ wchar_t* rhash_wcsdup(const wchar_t* str, const char* srcfile, int srcline)
 	wchar_t* res = (wchar_t*)malloc((wcslen(str) + 1) * sizeof(wchar_t));
 	if (res) wcscpy(res, str);
 #endif
-
-	if (!res) {
-		rsh_report_error(srcfile, srcline, "wcsdup(\"%u\") failed\n", (wcslen(str) + 1));
-		rsh_exit(2);
-	}
+	RSH_REQUIRE_IMPL(res, srcfile, srcline, "wcsdup(\"%u\") failed\n", (wcslen(str) + 1));
 	return res;
 }
 #endif
@@ -461,13 +422,32 @@ wchar_t* rhash_wcsdup(const wchar_t* str, const char* srcfile, int srcline)
 void* rhash_realloc(void* mem, size_t size, const char* srcfile, int srcline)
 {
 	void* res = realloc(mem, size);
-	if (!res) {
-		rsh_report_error(srcfile, srcline, "realloc(%p, %u) failed\n", mem, (unsigned)size);
-		rsh_exit(2);
-	}
+	RSH_REQUIRE_IMPL(res, srcfile, srcline, "realloc(%p, %u) failed\n", mem, (unsigned)size);
 	return res;
 }
 
+/*=========================================================================
+ * Bit functions
+ *=========================================================================*/
+ #ifndef get_ctz
+/**
+ * Returns index of the trailing bit of a 32-bit number.
+ * This is a plain C equivalent for GCC __builtin_ctz() bit scan.
+ *
+ * @param x the number to process
+ * @return zero-based index of the trailing bit
+ */
+unsigned get_ctz(unsigned x)
+{
+	/* array for conversion to bit position */
+	static unsigned char bit_pos[32] =  {
+		0, 1, 28, 2, 29, 14, 24, 3, 30, 22, 20, 15, 25, 17, 4, 8,
+		31, 27, 13, 23, 21, 19, 16, 7, 26, 12, 18, 6, 11, 5, 10, 9
+	};
+	return bit_pos[((uint32_t)((x & -x) * 0x077CB531U)) >> 27];
+}
+#endif
+
 /*=========================================================================
  * Containers
  *=========================================================================*/
