@@ -53,7 +53,7 @@ static int str_is_ascii(const char* str)
  * @param path file path
  * @return file name
  */
-static const char* get_basename(const char* path)
+const char* get_basename(const char* path)
 {
 	const char* p;
 	if (!path)
@@ -62,37 +62,48 @@ static const char* get_basename(const char* path)
 	return p;
 }
 
+#ifdef _WIN32
+# define CHECK_SLASH(ch, alt_sep) IS_ANY_SLASH(ch)
+#else
+# define CHECK_SLASH(ch, alt_sep) ((ch) == SYS_PATH_SEPARATOR || (ch) == (alt_sep))
+#endif
+
 /**
  * Return filepath, obtained by concatinating a directory path and a sub-path.
  *
  * @param dir_path (nullable) directory path
  * @param sub_path the filepath to append to the directory
- * @param user_path_separator flag, 1 to use user-defined path separator,
+ * @param flags bitmask consisting of the following bits:
+ *        FileInitUpdatePrintPathSlashes - contacenate using user-defined slash
+ *        FileInitUseRealPathAsIs - do not replace alternative path separators
  *        0 to use system path separator
  * @return concatinated file path
  */
-char* make_path(const char* dir_path, const char* sub_path, int user_path_separator)
+char* make_path(const char* dir_path, const char* sub_path, unsigned flags)
 {
 	char* buf;
 	size_t dir_len;
+#ifndef _WIN32
+	char alt_sep = (flags & FileInitUseRealPathAsIs ? SYS_PATH_SEPARATOR : ALIEN_PATH_SEPARATOR);
+#endif
 	assert(sub_path);
-	if (sub_path[0] == '.' && IS_ANY_SLASH(sub_path[1]))
+	if (sub_path[0] == '.' && CHECK_SLASH(sub_path[1], alt_sep))
 		sub_path += 2;
 	if (!dir_path)
 		return rsh_strdup(sub_path);
 	/* remove leading path delimiters from sub_path */
-	for (; IS_ANY_SLASH(*sub_path); sub_path++);
+	for (; CHECK_SLASH(*sub_path, alt_sep); sub_path++);
 	if (dir_path[0] == 0 || IS_DOT_STR(dir_path)) {
 		/* do not extend sub_path for dir_path="." */
 		return rsh_strdup(sub_path);
 	}
 	/* remove trailing path delimiters from the directory path */
-	for (dir_len = strlen(dir_path); dir_len > 0 && IS_ANY_SLASH(dir_path[dir_len - 1]); dir_len--);
+	for (dir_len = strlen(dir_path); dir_len > 0 && CHECK_SLASH(dir_path[dir_len - 1], alt_sep); dir_len--);
 	/* copy directory path */
 	buf = (char*)rsh_malloc(dir_len + strlen(sub_path) + 2);
 	memcpy(buf, dir_path, dir_len);
 	/* insert path separator */
-	buf[dir_len++] = (user_path_separator && opt.path_separator ? opt.path_separator : SYS_PATH_SEPARATOR);
+	buf[dir_len++] = ((flags & FileMaskUpdatePrintPath) && opt.path_separator ? opt.path_separator : SYS_PATH_SEPARATOR);
 	strcpy(buf + dir_len, sub_path); /* append sub_path */
 	return buf;
 }
@@ -126,7 +137,7 @@ tpath_t make_wpath(ctpath_t dir_path, size_t dir_len, ctpath_t sub_path)
 	result = (wchar_t*)rsh_malloc((dir_len + len + 2) * sizeof(wchar_t));
 	if (dir_len > 0) {
 		memcpy(result, dir_path, dir_len * sizeof(wchar_t));
-		if (result[dir_len - 1] != L'\\' && sub_path[0]) {
+		if (!IS_PATH_SEPARATOR_W(result[dir_len - 1]) && sub_path[0]) {
 			/* append path separator to the directory */
 			result[dir_len++] = L'\\';
 		}
@@ -196,24 +207,6 @@ static void convert_backslashes_to_unix(char* path)
 }
 #endif /* _WIN32 */
 
-/**
- * Check if a path points to a regular file.
- *
- * @param path the path to check
- * @return 1 if file exists an is a regular file, 0 otherwise
- */
-int is_regular_file(const char* path)
-{
-	int is_regular = 0;
-	file_t file;
-	file_init_by_print_path(&file, NULL, path, FileInitReusePath);
-	if (file_stat(&file, 0) >= 0) {
-		is_regular = FILE_ISREG(&file);
-	}
-	file_cleanup(&file);
-	return is_regular;
-}
-
 /*=========================================================================
  * file_t functions
  *=========================================================================*/
@@ -336,7 +329,7 @@ static int detect_path_encoding(file_t* file, wchar_t* dir_path, const char* pri
 			}
 			free(file->real_path);
 			file->real_path = last_path;
-			if(file->real_path)
+			if (file->real_path)
 				return primary_path_index;
 		} else if (file->real_path) {
 			return (path_index & 1);
@@ -359,6 +352,9 @@ static int detect_path_encoding(file_t* file, wchar_t* dir_path, const char* pri
  */
 int file_init_by_print_path(file_t* file, file_t* prepend_dir, const char* print_path, unsigned init_flags)
 {
+#ifndef _WIN32
+	char alt_sep = (init_flags & FileInitUseRealPathAsIs ? SYS_PATH_SEPARATOR : ALIEN_PATH_SEPARATOR);
+#endif
 	assert(print_path);
 	assert(!prepend_dir || prepend_dir->real_path);
 	memset(file, 0, sizeof(file_t));
@@ -368,7 +364,7 @@ int file_init_by_print_path(file_t* file, file_t* prepend_dir, const char* print
 		file->mode |= FileDontFreePrintPath | FileIsAsciiPrintPath;
 		return 0;
 	}
-	if (print_path[0] == '.' && IS_PATH_SEPARATOR(print_path[1]))
+	if (print_path[0] == '.' && CHECK_SLASH(print_path[1], alt_sep))
 		print_path += 2;
 #ifdef _WIN32
 	{
@@ -391,9 +387,9 @@ int file_init_by_print_path(file_t* file, file_t* prepend_dir, const char* print
 	}
 #else
 	if (!prepend_dir || IS_DOT_STR(prepend_dir->real_path)) {
-		file_init(file, print_path, init_flags & (FileInitReusePath | FileMaskModeBits));
+		file_init(file, print_path, init_flags & (FileInitReusePath | FileInitUseRealPathAsIs | FileMaskModeBits));
 	} else {
-		file->real_path = make_path(prepend_dir->real_path, print_path, 0);
+		file->real_path = make_path(prepend_dir->real_path, print_path, init_flags & FileInitUseRealPathAsIs);
 		file->mode = init_flags & FileMaskModeBits;
 	}
 	assert(file->print_path == NULL);
@@ -412,14 +408,33 @@ int file_init_by_print_path(file_t* file, file_t* prepend_dir, const char* print
 	return 0;
 }
 
+/**
+ * Restore a character of the file path, if it was saved earler.
+ *
+ * @param file the file
+ */
+static void restore_saved_char(file_t* file)
+{
+	unsigned saved = file->mode & FileMaskSavedChar;
+	assert(((unsigned)(FileMaskSavedChar) >> 24) == 0xff);
+	if (saved) {
+		file->mode &= ~FileMaskSavedChar;
+		if (file->data) {
+			*(file->data) = (char)(unsigned char)(saved >> 24);
+			file->data = 0;
+		}
+	}
+}
+
 /**
  * Transform the given file path, according to passed flags.
  *
+ * @param file the file
  * @param path the file path to transform
  * @param flags bitmask containing FPathBaseName, FPathNotNull and FileMaskUpdatePrintPath bit flags
  * @return transformed path
  */
-static const char* handle_rest_of_path_flags(const char* path, unsigned flags)
+static const char* handle_rest_of_path_flags(file_t* file, const char* path, unsigned flags)
 {
 	if (path == NULL)
 		return ((flags & FPathNotNull) ? (errno == EINVAL ? "(null)" : "(encoding error)") : NULL);
@@ -433,7 +448,24 @@ static const char* handle_rest_of_path_flags(const char* path, unsigned flags)
 			}
 		}
 	}
-	return (flags & FPathBaseName ? get_basename(path) : path);
+	if ((flags & (FPathBaseName | FPathDirName)) != 0)
+	{
+		const char* name = get_basename(path);
+		if ((flags & FPathBaseName) != 0)
+			return name;
+		if (name <= path)
+			return ".";
+		name--; /* point to the end of the file dirname */
+		if (*name && file)
+		{
+			/* store the character that terminates dirname */
+			assert((unsigned)FileMaskSavedChar == (unsigned)(0xff << 24));
+			file->mode |= ((unsigned)*name) << 24;
+			file->data = (char*)name;
+			*(char*)name = 0;
+		}
+	}
+	return path;
 }
 
 /**
@@ -456,8 +488,9 @@ const char* file_get_print_path(file_t* file, unsigned flags)
 	int is_utf8 = (opt.flags & OPT_UTF8 ? !(flags & FPathNative) : flags & FPathUtf8);
 	const char* secondary_path;
 	const char** primary_path = (is_utf8 || (file->mode & FileIsAsciiPrintPath) ? &file->print_path : &file->native_path);
+	restore_saved_char(file);
 	if (*primary_path)
-		return handle_rest_of_path_flags(*primary_path, flags);
+		return handle_rest_of_path_flags(file, *primary_path, flags);
 	if (is_utf8) {
 		convert_to = ConvertToUtf8;
 		dont_use_bit = FileDontUsePrintPath;
@@ -474,15 +507,15 @@ const char* file_get_print_path(file_t* file, unsigned flags)
 				file->mode |= dont_use_bit;
 		} else
 			errno = EILSEQ;
-		return handle_rest_of_path_flags(*primary_path, flags);
+		return handle_rest_of_path_flags(file, *primary_path, flags);
 	}
 	if (!file->real_path) {
 		errno = EINVAL;
-		return handle_rest_of_path_flags(NULL, flags);
+		return handle_rest_of_path_flags(NULL, NULL, flags);
 	}
 	*primary_path = convert_wcs_to_str(file->real_path, convert_to | ConvertPath);
 	if (!*primary_path)
-		return handle_rest_of_path_flags(NULL, flags);
+		return handle_rest_of_path_flags(NULL, NULL, flags);
 	if (str_is_ascii(*primary_path)) {
 		file->mode |= FileIsAsciiPrintPath;
 		if (primary_path != &file->print_path) {
@@ -491,13 +524,21 @@ const char* file_get_print_path(file_t* file, unsigned flags)
 			primary_path = &file->print_path;
 		}
 	}
-	return handle_rest_of_path_flags(*primary_path, flags);
+	return handle_rest_of_path_flags(file, *primary_path, flags);
 #else
+	const char* path = NULL;
+	restore_saved_char(file);
 	if (!file->print_path && !file->real_path)
 		errno = EINVAL;
-	if (!file->print_path && (flags & FileMaskUpdatePrintPath))
-		file->print_path = rsh_strdup(file->real_path);
-	return handle_rest_of_path_flags((file->print_path ? file->print_path : file->real_path), flags);
+	else if ((flags & FPathReal) != 0) {
+		path = (file->real_path ? file->real_path : file->print_path);
+		flags &= ~FileMaskUpdatePrintPath;
+	} else {
+		if (!file->print_path && (flags & FileMaskUpdatePrintPath))
+			file->print_path = rsh_strdup(file->real_path);
+		path = (file->print_path ? file->print_path : file->real_path);
+	}
+	return handle_rest_of_path_flags(file, path, flags);
 #endif
 }
 
@@ -521,7 +562,8 @@ void file_cleanup(file_t* file)
 	file->native_path = NULL;
 #endif /* _WIN32 */
 
-	free(file->data);
+	if (file->mode & FileIsData)
+		free(file->data);
 	file->data = NULL;
 	file->mtime = 0;
 	file->size = 0;
@@ -689,17 +731,34 @@ static int file_statw(file_t* file)
 
 	/* read file attributes */
 	if (GetFileAttributesExW(file->real_path, GetFileExInfoStandard, &data)) {
-		uint64_t u;
-		file->size  = (((uint64_t)data.nFileSizeHigh) << 32) + data.nFileSizeLow;
 		file->mode |= (data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ? FileIsDir : FileIsReg);
-		if ((data.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) != 0)
-			file->mode |= FileIsLnk;
+		do {
+			uint64_t u;
+			if ((data.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) == 0) {
+				/* Process a regular file or directory */
+				file->size  = (((uint64_t)data.nFileSizeHigh) << 32) + data.nFileSizeLow;
+			} else {
+				/* Process a symlink (aka Reparse point) */
+				LARGE_INTEGER link_size;
+				HANDLE handle = CreateFileW(file->real_path,
+					GENERIC_READ, FILE_SHARE_READ, NULL,
+					OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
+				if (handle == INVALID_HANDLE_VALUE)
+					break;
+				if (!GetFileSizeEx(handle, &link_size))
+					break;
+				if (!GetFileTime(handle, NULL, NULL, &data.ftLastWriteTime))
+					break;
+				file->size  = (((uint64_t)link_size.HighPart) << 32) + link_size.LowPart;
+				file->mode |= FileIsLnk;
+			}
 
-		/* the number of 100-nanosecond intervals since January 1, 1601 */
-		u = (((uint64_t)data.ftLastWriteTime.dwHighDateTime) << 32) + data.ftLastWriteTime.dwLowDateTime;
-		/* convert to seconds and subtract the epoch difference */
-		file->mtime = u / 10000000 - 11644473600LL;
-		return 0;
+			/* the number of 100-nanosecond intervals since January 1, 1601 */
+			u = (((uint64_t)data.ftLastWriteTime.dwHighDateTime) << 32) + data.ftLastWriteTime.dwLowDateTime;
+			/* convert to seconds and subtract the epoch difference */
+			file->mtime = u / 10000000 - 11644473600LL;
+			return 0;
+		} while (0);
 	}
 	file->mode |= FileIsInaccessible;
 	set_errno_from_last_file_error();
@@ -791,7 +850,7 @@ FILE* file_fopen(file_t* file, int fopen_flags)
 #else
 	fd = fopen(file->real_path, mode);
 # if _POSIX_C_SOURCE >= 200112L && !defined(__STRICT_ANSI__)
-	if(fd)
+	if (fd)
 		posix_fadvise(fileno(fd), 0, 0, POSIX_FADV_SEQUENTIAL);
 # endif /* _POSIX_C_SOURCE >= 200112L && !defined(__STRICT_ANSI__) */
 	return fd;
@@ -937,7 +996,7 @@ int file_list_read(file_list_t* list)
 {
 	char buf[2048];
 	file_cleanup(&list->current_file);
-	while(fgets(buf, 2048, list->fd)) {
+	while (fgets(buf, 2048, list->fd)) {
 		char* p;
 		char* line = buf;
 		char* buf_back = buf + sizeof(buf) - 1;
