@@ -107,12 +107,13 @@ static size_t unescape_characters(char* buffer, size_t buffer_size, const char*
 	for (; src < src_end && dst < dst_back; dst++) {
 		*dst = *(src++); /* copy non-escaped characters */
 		if (*dst == '\\') {
-			if (*src == '\\') {
-				src++; /* interpret '\\' as a single '\' */
-			} else if (*src == 'n') {
+			if (*src == 'r')
+				*dst = '\r';
+			else if (*src == 'n')
 				*dst = '\n';
-				src++;
-			}
+			else if (*src != '\\')
+				continue;
+			src++;
 		}
 	}
 	assert(dst <= dst_back);
@@ -120,29 +121,6 @@ static size_t unescape_characters(char* buffer, size_t buffer_size, const char*
 	return (dst - buffer);
 }
 
-/* convert a hash function bit-flag to the index of the bit */
-#if __GNUC__ >= 4 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4) /* GCC >= 3.4 */
-# define get_ctz(x) __builtin_ctz(x)
-#else
-
-/**
- * Returns index of the trailing bit of a 32-bit number.
- * This is a plain C equivalent for GCC __builtin_ctz() bit scan.
- *
- * @param x the number to process
- * @return zero-based index of the trailing bit
- */
-static unsigned get_ctz(unsigned x)
-{
-	/* see http://graphics.stanford.edu/~seander/bithacks.html */
-	static unsigned char bit_pos[32] =  {
-		0, 1, 28, 2, 29, 14, 24, 3, 30, 22, 20, 15, 25, 17, 4, 8,
-		31, 27, 13, 23, 21, 19, 16, 7, 26, 12, 18, 6, 11, 5, 10, 9
-	};
-	return bit_pos[((uint32_t)((x & -(int)x) * 0x077CB531U)) >> 27];
-}
-#endif /* (GCC >= 4.3) */
-
 /**
  * Encode a hash function digest size into a small number in [0,...,7].
  * The digest size must be in the set { 4, 16, 20, 24, 28, 32, 48, 64 }.
@@ -160,27 +138,31 @@ static int code_digest_size(int digest_size)
 }
 
 /**
- * Calculate a bit-mask of hash-ids by a length of message digest in bytes.
+ * Compute a bit mask for hash functions that have a specified length
+ * (in bytes) of their message digest.
  *
- * @param digest_size length of a binary message digest in bytes
- * @return mask of hash-ids with given hash length, 0 on fail
+ * @param digest_size length (in bytes) of a binary message digest
+ * @return bit mask for found hash functions, 0 on fail
  */
-static unsigned hash_bitmask_by_digest_size(int digest_size)
+static uint64_t hash_bitmask_by_digest_size(int digest_size)
 {
-	static unsigned mask[10] = { 0,0,0,0,0,0,0,0,0,0 };
+	static uint64_t hash_masks[10] = { 0,0,0,0,0,0,0,0,0,0 };
 	int code;
-	if (mask[9] == 0) {
-		unsigned hash_id;
-		for (hash_id = 1; hash_id <= RHASH_ALL_HASHES; hash_id <<= 1) {
-			code = code_digest_size(rhash_get_digest_size(hash_id));
+	if (hash_masks[9] == 0) {
+		unsigned hash_ids[64];
+		size_t count = rhash_get_all_algorithms(64, hash_ids);
+		size_t i;
+		RSH_REQUIRE(count != RHASH_ERROR, "failed to get all hash ids\n");
+		for (i = 0; i < count; i++) {
+			code = code_digest_size(rhash_get_digest_size(hash_ids[i]));
 			assert(0 <= code && code <= 7);
 			if (code >= 0)
-				mask[code] |= hash_id;
+				hash_masks[code] |= hash_id_to_bit64(hash_ids[i]);
 		}
-		mask[9] = 1;
+		hash_masks[9] = 1;
 	}
 	code = code_digest_size(digest_size);
-	return (code >= 0 ? mask[code] : 0);
+	return (code >= 0 ? hash_masks[code] : 0);
 }
 
 /**
@@ -482,7 +464,7 @@ static int match_hash_tokens(struct hash_token* token, const char* format, unsig
 	char* begin = token->begin;
 	char* end = token->end;
 	char* url_path = 0;
-	size_t url_length;
+	size_t url_length = 0;
 	struct hash_parser* hp = &token->parser->hp;
 	struct hash_value hv;
 	int unsaved_hashval = 0;
@@ -544,9 +526,17 @@ static int match_hash_tokens(struct hash_token* token, const char* format, unsig
 		case '\2':
 		case '\3':
 			if (backward) {
+				/* trim right */
+				for (; begin < end && rhash_isspace(end[-1]); end--, len++);
+				if (begin == end)
+					return ResFailed;
 				hv.format = test_hash_string(&end, begin, &len);
 				hv.offset = (unsigned short)(end - hp->line_begin);
 			} else {
+				/* trim left */
+				for (; rhash_isspace(*begin) && begin < end; begin++, len++);
+				if (begin == end)
+					return ResFailed;
 				hv.offset = (unsigned short)(begin - hp->line_begin);
 				hv.format = test_hash_string(&begin, end, &len) & token->hash_type;
 			}
@@ -564,8 +554,8 @@ static int match_hash_tokens(struct hash_token* token, const char* format, unsig
 					hv.format &= ~FmtBase64;
 				if (!hv.format)
 					return ResFailed;
-				hv.hash_id = token->expected_hash_id;
-			} else hv.hash_id = 0;
+				hv.hash_mask = hash_id_to_bit64(token->expected_hash_id);
+			} else hv.hash_mask = 0;
 			hv.length = (unsigned char)len;
 			unsaved_hashval = 1;
 			break;
@@ -591,12 +581,15 @@ static int match_hash_tokens(struct hash_token* token, const char* format, unsig
 			break;
 		case '\6':
 		case ' ':
-			if (backward)
-				for (; begin < end && rhash_isspace(end[-1]); end--, len++);
-			else
-				for (; rhash_isspace(*begin) && begin < end; begin++, len++);
+			if (backward) {
+				char* low_limit = (begin < end && *search_str == '\6' ? end - 1 : begin);
+				for (; low_limit < end && rhash_isspace(end[-1]); end--, len++);
+			} else {
+				char* hi_limit = (begin < end && *search_str == '\6' ? begin + 1 : end);
+				for (; begin < hi_limit && rhash_isspace(*begin); begin++, len++);
+			}
 			/* check if space is mandatory */
-			if (*search_str != ' ' && len == 0) {
+			if (*search_str == '\6' && len == 0) {
 				/* for '\6' verify (len > 0 || (MhtAllowOneHash is set && begin == end)) */
 				if (!(bit_flags & MhtAllowOneHash) || begin < end)
 					return ResFailed;
@@ -614,17 +607,23 @@ static int match_hash_tokens(struct hash_token* token, const char* format, unsig
 	if (unsaved_hashval && hp->hashes_num < HP_MAX_HASHES) {
 		token->p_hashes[hp->hashes_num++] = hv;
 	}
+	if (!backward) {
+		/* eat zero or one leading space */
+		if (begin < end && rhash_isspace(*begin))
+			begin++;
+	}
+
+	/* the rest is stored as a file path */
 	token->begin = begin;
 	token->end = end;
 	if ((bit_flags & MhtAllowOneHash) != 0 && hp->hashes_num == 1 && begin == end)
 	{
+		/* a single hash without a file path was detected */
 		struct hash_parser_ext* const parser = token->parser;
 		file_t* parsed_path = &parser->hp.parsed_path;
 
-		if (file_modify_path(parsed_path, parser->hash_file, NULL, FModifyRemoveExtension) < 0) {
-			/* note: trailing whitespaces were removed from line by hash_parser_parse_line() */
+		if (file_modify_path(parsed_path, parser->hash_file, NULL, FModifyRemoveExtension) < 0)
 			return ResFailed;
-		}
 		if ((bit_flags & MhtFstatPath) != 0 && file_stat(parsed_path, 0) < 0)
 			hp->parsed_path_errno = errno;
 		return ResOneHashDetected;
@@ -714,25 +713,25 @@ static int finalize_parsed_data(struct hash_token* token)
 		char* hash_str = parser->line_begin + hv->offset;
 		hash_str[hv->length] = '\0'; /* terminate the message digest */
 
-		if (hv->hash_id == 0) {
-			/* calculate bit-mask of hash function ids */
-			unsigned mask = 0;
+		if (hv->hash_mask == 0) {
+			/* calculate bit-mask of hash function */
+			uint64_t hash_mask = 0;
 			if (hv->format & FmtHex) {
-				mask |= hash_bitmask_by_digest_size(hv->length >> 1);
+				hash_mask |= hash_bitmask_by_digest_size(hv->length >> 1);
 			}
 			if (hv->format & FmtBase32) {
 				assert(((hv->length * 5 / 8) & 3) == 0);
-				mask |= hash_bitmask_by_digest_size(BASE32_BIT_SIZE(hv->length) / 8);
+				hash_mask |= hash_bitmask_by_digest_size(BASE32_BIT_SIZE(hv->length) / 8);
 			}
 			if (hv->format & FmtBase64) {
-				mask |= hash_bitmask_by_digest_size(BASE64_BIT_SIZE(hv->length) / 8);
+				hash_mask |= hash_bitmask_by_digest_size(BASE64_BIT_SIZE(hv->length) / 8);
 			}
-			assert(mask != 0);
-			if ((mask & token->parser->expected_hash_mask) != 0)
-				mask &= token->parser->expected_hash_mask;
-			hv->hash_id = mask;
+			assert(hash_mask != 0);
+			if ((hash_mask & token->parser->expected_hash_mask) != 0)
+				hash_mask &= token->parser->expected_hash_mask;
+			hv->hash_mask = hash_mask;
 		}
-		parser->hash_mask |= hv->hash_id;
+		parser->hash_mask |= hv->hash_mask;
 	}
 	return ResAllMatched;
 }
@@ -786,7 +785,7 @@ static int parse_magnet_url(struct hash_token* token)
 							break;
 					}
 					if (i >= RHASH_HASH_COUNT) {
-						if (opt.flags & OPT_VERBOSE) {
+						if (opt.verbose) {
 							*hf_end = '\0';
 							log_warning(_("unknown hash in magnet link: %s\n"), token->begin);
 						}
@@ -853,35 +852,36 @@ static int parse_ed2k_link(struct hash_token* token)
 static int parse_hash_file_line(struct hash_parser_ext* parser, int check_eol)
 {
 	struct hash_token token;
-	char* line = parser->hp.line_begin;
-	char* line_end = strchr(line, '\0');
+	char* const line_start = parser->hp.line_begin;
+	char* token_start = line_start;
+	char* line_end = strchr(line_start, '\0');
 	int res;
 
-	assert(line[0] != '\0');
+	assert(line_start[0] != '\0');
 
 	/* return if EOL not found at the end of the line */
 	if (line_end[-1] != '\n' && check_eol)
 		return ResFailed;
-	/* trim line manually, without str_trim(), cause we'll need line_end later */
-	while (rhash_isspace(line_end[-1]) && line_end > line)
+	/* remove trailing CR and LF */
+	while ((line_end[-1] == '\n' || line_end[-1] == '\r') && line_end > line_start)
 		*(--line_end) = 0;
 	/* skip white spaces at the start of the line */
-	while (rhash_isspace(*line))
-		line++;
+	while (rhash_isspace(*token_start))
+		token_start++;
 
 	memset(&token, 0, sizeof(token));
-	token.begin = line;
+	token.begin = token_start;
 	token.end = line_end;
 	token.parser = parser;
 	token.p_parsed_path = &parser->hp.parsed_path;
 	token.p_hashes = parser->hp.hashes;
 
 	memset(&parser->hp, 0, sizeof(parser->hp));
-	parser->hp.line_begin = line;
+	parser->hp.line_begin = line_start;
 	parser->hp.file_size = (uint64_t)-1;
 
 	/* check for a minimum acceptable message digest length */
-	if ((line + 8) > line_end)
+	if ((token_start + 8) > line_end)
 		return ResFailed;
 
 	if (strncmp(token.begin, "ed2k://|file|", 13) == 0) {
@@ -893,10 +893,11 @@ static int parse_hash_file_line(struct hash_parser_ext* parser, int check_eol)
 		return parse_magnet_url(&token);
 	}
 	/* check for BSD-formatted line has been processed */
-	res = match_hash_tokens(&token, "\1 ( $ ) = \3", MhtFstatPath | MhtAllowEscapedPath);
+	res = match_hash_tokens(&token, "\1 ($) = \3", MhtFstatPath | MhtAllowEscapedPath);
 	if (res != ResFailed)
 		return finalize_parsed_data(&token);
 	token.hash_type = FmtAll;
+	token.begin = line_start;
 
 	{
 		const unsigned is_sfv_format = parser->is_sfv;
@@ -1038,11 +1039,13 @@ unsigned get_crc32(struct rhash_context* ctx)
  */
 static int do_hash_sums_match(struct hash_parser* parser, struct rhash_context* ctx)
 {
+	uint64_t hash_mask = parser->hash_mask;
 	unsigned unverified_mask;
-	unsigned hash_id;
 	unsigned printed;
 	char hex[132], base32[104], base64[88];
 	int j;
+	static const uint64_t gost94_hash_mask =
+		hash_id_to_bit64(RHASH_GOST94) | hash_id_to_bit64(RHASH_GOST94_CRYPTOPRO);
 
 	/* verify file size, if present */
 	if ((parser->bit_flags & HpHasFileSize) != 0 && parser->file_size != ctx->msg_size)
@@ -1058,8 +1061,10 @@ static int do_hash_sums_match(struct hash_parser* parser, struct rhash_context*
 
 	unverified_mask = (1 << parser->hashes_num) - 1;
 
-	for (hash_id = 1; hash_id <= RHASH_ALL_HASHES && unverified_mask; hash_id <<= 1) {
-		if ((parser->hash_mask & hash_id) == 0)
+	while(hash_mask && unverified_mask) {
+		uint64_t bit64 = hash_mask & -hash_mask;
+		unsigned hash_id = bit64_to_hash_id(bit64);
+		if ((parser->hash_mask & bit64) == 0)
 			continue;
 		printed = 0;
 
@@ -1071,7 +1076,7 @@ static int do_hash_sums_match(struct hash_parser* parser, struct rhash_context*
 			int comparision_mode;
 
 			/* skip already verified message digests and message digests of different size */
-			if (!(unverified_mask & (1 << j)) || !(hv->hash_id & hash_id))
+			if (!(unverified_mask & (1 << j)) || !(hv->hash_mask & bit64))
 				continue;
 			comparision_mode = 0;
 			bit_length = rhash_get_digest_size(hash_id) * 8;
@@ -1085,7 +1090,7 @@ static int do_hash_sums_match(struct hash_parser* parser, struct rhash_context*
 					printed |= FmtHex;
 				}
 				calculated_hash = hex;
-				if ((hash_id & (RHASH_GOST94 | RHASH_GOST94_CRYPTOPRO)) != 0)
+				if ((bit64 & gost94_hash_mask) != 0)
 					comparision_mode = CompareHashReversed;
 			} else if (hv->length == BASE32_LENGTH(bit_length)) {
 				assert(hv->format & FmtBase32);
@@ -1115,12 +1120,13 @@ static int do_hash_sums_match(struct hash_parser* parser, struct rhash_context*
 				continue;
 
 			unverified_mask &= ~(1 << j); /* mark the j-th message digest as verified */
-			parser->found_hash_ids |= hash_id;
+			parser->found_hash_ids |= bit64;
 
 			/* end the loop if all message digests were successfully verified */
 			if (unverified_mask == 0)
 				break;
 		}
+		hash_mask ^= bit64;
 	}
 
 	parser->wrong_hashes = unverified_mask;
@@ -1149,8 +1155,9 @@ static int verify_hashes(file_t* file, struct hash_parser* hp)
 
 	memset(&info, 0, sizeof(info));
 	info.file = file;
-	info.sums_flags = hp->hash_mask;
+	info.hash_mask = hp->hash_mask;
 	info.hp = hp;
+	print_verbose_algorithms(rhash_data.log, info.hash_mask);
 
 	/* initialize percents output */
 	if (init_percents(&info) < 0) {
@@ -1186,7 +1193,7 @@ static int verify_hashes(file_t* file, struct hash_parser* hp)
 		res = 1;
 	if (finish_percents(&info, res) < 0)
 		res = -2;
-	if ((opt.flags & OPT_SPEED) && info.sums_flags != 0)
+	if ((opt.flags & OPT_SPEED) && info.hash_mask != 0)
 		print_file_time_stats(&info);
 	return res;
 }
@@ -1281,18 +1288,19 @@ static int extract_uppercase_file_ext(struct file_ext* fe, file_t* file)
 static void set_flags_by_hash_file_extension(file_t* hash_file, struct hash_parser_ext* parser)
 {
 	struct file_ext ext;
-	unsigned hash_mask;
+	uint64_t hash_mask;
 	int is_sfv;
-	if(HAS_OPTION(OPT_NO_DETECT_BY_EXT) || !extract_uppercase_file_ext(&ext, hash_file))
+	if (HAS_OPTION(OPT_NO_DETECT_BY_EXT) || !extract_uppercase_file_ext(&ext, hash_file))
 		return;
-	hash_mask = (opt.sum_flags ? opt.sum_flags : bsd_hash_name_to_id(ext.buffer, ext.length, PrefixMatch));
+	hash_mask = (opt.hash_mask ? opt.hash_mask :
+		hash_id_to_bit64(bsd_hash_name_to_id(ext.buffer, ext.length, PrefixMatch)));
 	is_sfv = (ext.length == 3 && memcmp(ext.buffer, "SFV", 3) == 0);
 	if (parser != NULL) {
 		parser->expected_hash_mask = hash_mask;
 		parser->is_sfv = is_sfv;
 	}
-	if (IS_MODE(MODE_UPDATE)) {
-		opt.sum_flags = hash_mask;
+	if (IS_MODE(MODE_UPDATE | MODE_MISSING | MODE_UNVERIFIED)) {
+		opt.hash_mask = hash_mask;
 		if (!opt.fmt)
 			rhash_data.is_sfv = is_sfv;
 	}
@@ -1341,13 +1349,13 @@ static struct hash_parser* hash_parser_open(file_t* hash_file, int chdir)
 	memset(parser, 0, sizeof(struct hash_parser_ext));
 	parser->hash_file = hash_file;
 	parser->fd = fd;
-	parser->expected_hash_mask = opt.sum_flags;
+	parser->expected_hash_mask = opt.hash_mask;
 	parser->is_sfv = rhash_data.is_sfv;
 
 	/* extract the parent directory of the hash file, if required */
 	if (chdir)
 		file_modify_path(&parser->parent_dir, hash_file, NULL, FModifyGetParentDir);
-	if((opt.flags & OPT_NO_DETECT_BY_EXT) == 0)
+	if ((opt.flags & OPT_NO_DETECT_BY_EXT) == 0)
 		set_flags_by_hash_file_extension(hash_file, parser);
 	return &parser->hp;
 }
@@ -1442,7 +1450,7 @@ static int hash_parser_process_file(struct hash_parser *parser, file_set* files)
 	rhash_data.processed = rhash_data.ok = rhash_data.miss = 0;
 	rhash_data.total_size = 0;
 
-	while((parsing_res = hash_parser_process_line(parser)) > ResStopParsing) {
+	while ((parsing_res = hash_parser_process_line(parser)) > ResStopParsing) {
 		result &= ~HashFileIsEmpty;
 		/* skip comments and empty lines */
 		if (parsing_res == ResSkipLine)
@@ -1484,6 +1492,18 @@ static int hash_parser_process_file(struct hash_parser *parser, file_set* files)
 					else
 						result |= HashFileHasWrongHashes;
 				}
+			} else if (IS_MODE(MODE_MISSING)) {
+				if (FILE_ISBAD(&parser->parsed_path)) {
+					/* print the missing file */
+					if (fprintf_file_t(rhash_data.out, "%s\n", &parser->parsed_path, OutEscapePrefixed) < 0) {
+						log_error_file_t(&rhash_data.out_file);
+						return -2;
+					}
+					result |= HashFileHasMissedFiles;
+					rhash_data.miss++;
+				}
+				else
+					rhash_data.ok++;
 			}
 		}
 		rhash_data.processed++;
@@ -1493,11 +1513,12 @@ static int hash_parser_process_file(struct hash_parser *parser, file_set* files)
 
 	time = rsh_timer_stop(&timer);
 
-	if (IS_MODE(MODE_CHECK)) {
+	if (IS_MODE(MODE_CHECK | MODE_MISSING)) {
 		/* check for a hash file errors */
 		if (result >= 0 && (result & (HashFileHasWrongHashes | HashFileHasMissedFiles | HashFileHasUnparsedLines)) != 0)
 			rhash_data.non_fatal_error = 1;
-
+	}
+	if (IS_MODE(MODE_CHECK)) {
 		if (result >= -1 && (print_verifying_footer() < 0 || print_check_stats() < 0)) {
 			log_error_file_t(&rhash_data.out_file);
 			result = -2;
