@@ -17,6 +17,14 @@
 # include <io.h>
 #endif /* _WIN32 */
 
+#ifdef _WIN32
+# define NEWLINE_STR "\r\n"
+# define CHARS_TO_ESCAPE "\r\n"
+#else
+# define NEWLINE_STR "\n"
+# define CHARS_TO_ESCAPE "\r\n\\"
+#endif
+
 /*=========================================================================
  * Formatted output functions and structures
  *=========================================================================*/
@@ -24,12 +32,14 @@
 /**
  * The table with information about hash functions.
  */
-print_hash_info hash_info_table[32];
+print_hash_info hash_info_table[RHASH_HASH_COUNT + 1];
 
 /**
  * Possible types of a print_item.
  */
 enum {
+	PRINT_FLAG_BASENAME = 0x40000,
+	PRINT_FLAG_DIRNAME = 0x80000,
 	PRINT_ED2K_LINK = 0x100000,
 	PRINT_FLAG_UPPERCASE = 0x200000,
 	PRINT_FLAG_RAW    = 0x0400000,
@@ -38,15 +48,21 @@ enum {
 	PRINT_FLAG_BASE64 = 0x2000000,
 	PRINT_FLAG_URLENCODE = 0x4000000,
 	PRINT_FLAG_PAD_WITH_ZERO = 0x8000000,
-	PRINT_FLAGS_ALL = PRINT_FLAG_UPPERCASE | PRINT_FLAG_PAD_WITH_ZERO | PRINT_FLAG_RAW
-		| PRINT_FLAG_HEX | PRINT_FLAG_BASE32 | PRINT_FLAG_BASE64 | PRINT_FLAG_URLENCODE,
+	PRINT_FLAGS_PATH_PARTS = PRINT_FLAG_BASENAME | PRINT_FLAG_DIRNAME,
+	PRINT_FLAGS_ALL = PRINT_FLAG_BASENAME | PRINT_FLAG_DIRNAME | PRINT_FLAG_UPPERCASE
+		| PRINT_FLAG_PAD_WITH_ZERO | PRINT_FLAG_RAW | PRINT_FLAG_HEX
+		| PRINT_FLAG_BASE32 | PRINT_FLAG_BASE64 | PRINT_FLAG_URLENCODE,
 	PRINT_STR = 0x10000000,
 	PRINT_ZERO,
 	PRINT_NEWLINE,
 	PRINT_FILEPATH,
-	PRINT_BASENAME,
 	PRINT_SIZE,
-	PRINT_MTIME /*PRINT_ATIME, PRINT_CTIME*/
+	PRINT_MTIME, /*PRINT_ATIME, PRINT_CTIME*/
+	PRINT_ESCAPE_LINE,
+	PRINT_BASENAME = PRINT_FILEPATH | PRINT_FLAG_BASENAME,
+	PRINT_DIRNAME = PRINT_FILEPATH | PRINT_FLAG_DIRNAME,
+	FLAG_HAS_MISC_PARTS = PRINT_FLAG_RAW,
+	FLAG_IS_COMPUTED = PRINT_FLAG_PAD_WITH_ZERO,
 };
 
 /**
@@ -56,7 +72,7 @@ typedef struct print_item
 {
 	struct print_item* next;
 	unsigned flags;
-	unsigned hash_id;
+	unsigned hash_mask;
 	unsigned width;
 	const char* data;
 } print_item;
@@ -68,15 +84,15 @@ static print_item* parse_percent_item(const char** str);
  * Allocate new print_item.
  *
  * @param flags the print_item flags
- * @param hash_id optional hash_id
+ * @param hash_mask optional hash_mask
  * @param data optional string to store
  * @return allocated print_item
  */
-static print_item* new_print_item(unsigned flags, unsigned hash_id, const char* data)
+static print_item* new_print_item(unsigned flags, uint64_t hash_mask, const char* data)
 {
 	print_item* item = (print_item*)rsh_malloc(sizeof(print_item));
 	item->flags = flags;
-	item->hash_id = hash_id;
+	item->hash_mask = hash_mask;
 	item->width = 0;
 	item->data = (data ? rsh_strdup(data) : NULL);
 	item->next = NULL;
@@ -133,19 +149,23 @@ static char parse_escaped_char(const char** pformat)
 /**
  * Parse format string.
  *
+ * @param format the format string to parse
+ * @param hash_mask bit mask for found hash functions
  * @return a print_item list with parsed information
  */
-print_item* parse_print_string(const char* format, unsigned* sum_mask)
+print_item* parse_print_string(const char* format, uint64_t* hash_mask)
 {
 	char* buf;
 	char* p;
 	print_item* list = NULL;
 	print_item* item = NULL;
 	print_item** tail;
+	print_item* esc_head = NULL;
+	print_item* esc_tail = NULL;
 
 	buf = p = (char*)rsh_malloc( strlen(format) + 1 );
 	tail = &list;
-	*sum_mask = 0;
+	*hash_mask = 0;
 
 	for (;;) {
 		while (*format && *format != '%' && *format != '\\')
@@ -153,16 +173,25 @@ print_item* parse_print_string(const char* format, unsigned* sum_mask)
 
 		if (*format == '\\') {
 			format++;
-			*p = parse_escaped_char(&format);
-			if (*p == '\0') {
-				item = new_print_item(PRINT_ZERO, 0, NULL);
-#ifdef _WIN32
-			} else if (*p == '\n') {
-				item = new_print_item(PRINT_NEWLINE, 0, NULL);
-#endif
+			if (*format == '^') {
+				format++;
+				item = new_print_item(PRINT_ESCAPE_LINE, 0, NULL);
+				esc_tail = item;
+				if (!esc_head)
+					esc_head = item;
 			} else {
-				p++;
-				continue;
+				unsigned cmd;
+				*p = parse_escaped_char(&format);
+				if (*p == '\0') {
+					cmd = PRINT_ZERO;
+				} else if (*p == '\n') {
+					cmd = PRINT_NEWLINE;
+					esc_tail = NULL;
+				} else {
+					p++;
+					continue;
+				}
+				item = new_print_item(cmd, 0, NULL);
 			}
 		} else if (*format == '%') {
 			if ( *(++format) == '%' ) {
@@ -174,8 +203,17 @@ print_item* parse_print_string(const char* format, unsigned* sum_mask)
 					*(p++) = '%';
 					continue;
 				}
-				if (item->hash_id)
-					*sum_mask |= item->hash_id;
+				if ((item->flags & ~(PRINT_FLAGS_ALL & ~PRINT_FLAG_URLENCODE)) == PRINT_FILEPATH && esc_tail) {
+					unsigned parts = (item->flags & PRINT_FLAGS_PATH_PARTS);
+					if (!parts)
+						parts = PRINT_FLAGS_PATH_PARTS;
+					esc_tail->flags |= parts;
+					assert(esc_head != NULL);
+					if (esc_head->flags != esc_tail->flags)
+						esc_head->flags |= FLAG_HAS_MISC_PARTS;
+				}
+				if (item->hash_mask)
+					*hash_mask |= item->hash_mask;
 			}
 		}
 		if (p > buf || (!*format && list == NULL && item == NULL)) {
@@ -196,6 +234,39 @@ print_item* parse_print_string(const char* format, unsigned* sum_mask)
 	return list;
 }
 
+/**
+ * Test file path for characters requiring escaping.
+ *
+ * @param file the filepath to test
+ * @param esc_flags flags of the PRINT_ESCAPE_LINE print_item
+ * @param path_flags bit mask, which can contain OutForceUtf8
+ * @return the result of the test for special characters in the file
+ *     dirname and basename, represented as bit mask containing flags
+ *     PRINT_FLAG_BASENAME, PRINT_FLAG_DIRNAME, FLAG_IS_COMPUTED
+ */
+static unsigned get_file_escaping_flags(file_t* file, unsigned esc_flags, unsigned path_flags)
+{
+	const unsigned parts_flags = esc_flags & (FLAG_HAS_MISC_PARTS | PRINT_FLAGS_PATH_PARTS);
+	if (parts_flags && !FILE_ISSPECIAL(file)) {
+		const char* path = file_get_print_path(file, path_flags | FPathReal);
+		if (path && path[0]) {
+			const char* basename = get_basename(path);
+			const char* start = (parts_flags == PRINT_FLAG_BASENAME ? basename : path);
+			size_t pos = strcspn(start, CHARS_TO_ESCAPE);
+			if (start[pos]) {
+				if ((start + pos) >= basename)
+					return PRINT_FLAG_BASENAME | FLAG_IS_COMPUTED;
+				if (parts_flags == PRINT_FLAG_DIRNAME)
+					return PRINT_FLAG_DIRNAME | FLAG_IS_COMPUTED;
+				pos = strcspn(basename, CHARS_TO_ESCAPE);
+				return (!basename[pos] ? PRINT_FLAG_DIRNAME | FLAG_IS_COMPUTED :
+					PRINT_FLAG_BASENAME | PRINT_FLAG_DIRNAME | FLAG_IS_COMPUTED);
+			}
+		}
+	}
+	return FLAG_IS_COMPUTED;
+}
+
 /**
  * Convert given case-insensitive name to a printf directive id
  *
@@ -209,10 +280,11 @@ static unsigned printf_name_to_id(const char* name, size_t length, unsigned* fla
 	char buf[20];
 	size_t i;
 	print_hash_info* info = hash_info_table;
-	unsigned bit;
 
-	if (length > (sizeof(buf) - 1)) return 0;
-	for (i = 0; i < length; i++) buf[i] = tolower(name[i]);
+	if (length > (sizeof(buf) - 1))
+		return 0;
+	for (i = 0; i < length; i++)
+		buf[i] = tolower(name[i]);
 
 	/* check for legacy '%{urlname}' directive for compatibility */
 	if (length == 7 && memcmp(buf, "urlname", 7) == 0) {
@@ -223,9 +295,11 @@ static unsigned printf_name_to_id(const char* name, size_t length, unsigned* fla
 		return 0;
 	}
 
-	for (bit = 1; bit <= RHASH_ALL_HASHES; bit = bit << 1, info++) {
+	/* loop by hash functions */
+	for (info = hash_info_table; info->hash_id; info++) {
 		if (memcmp(buf, info->short_name, length) == 0 &&
-			info->short_name[length] == 0) return bit;
+				info->short_name[length] == 0)
+			return info->hash_id;
 	}
 	return 0;
 }
@@ -240,21 +314,21 @@ print_item* parse_percent_item(const char** str)
 {
 	const char* format = *str;
 	const char* p = NULL;
-	unsigned hash_id = 0;
+	uint64_t hash_mask = 0;
 	unsigned modifier_flags = 0;
 	int id_found = 0;
 	int width = 0;
 	print_item* item = NULL;
 
 	static const char* short_hash = "CMHTGWRAE";
-	static const char* short_other = "Llpfs";
+	static const char* short_other = "Llpfds";
 	static const unsigned hash_ids[] = {
 		RHASH_CRC32, RHASH_MD5, RHASH_SHA1, RHASH_TTH, RHASH_GOST12_256,
 		RHASH_WHIRLPOOL, RHASH_RIPEMD160, RHASH_AICH, RHASH_ED2K
 	};
 	static const unsigned other_flags[] = {
 		(PRINT_ED2K_LINK | PRINT_FLAG_UPPERCASE), PRINT_ED2K_LINK,
-		PRINT_FILEPATH, PRINT_BASENAME, PRINT_SIZE
+		PRINT_FILEPATH, PRINT_BASENAME, PRINT_DIRNAME, PRINT_SIZE
 	};
 	/* detect the padding by zeros */
 	if (*format == '0') {
@@ -288,8 +362,8 @@ print_item* parse_percent_item(const char** str)
 		for (; isalnum((unsigned char)*p) || (*p == '-'); p++);
 		if (*p == '}') {
 			unsigned flags = 0;
-			hash_id = printf_name_to_id(format + 1, p - (format + 1), &flags);
-			if (hash_id || (flags & PRINT_FLAG_URLENCODE) || flags == PRINT_MTIME) {
+			hash_mask = printf_name_to_id(format + 1, p - (format + 1), &flags);
+			if (hash_mask || (flags & PRINT_FLAG_URLENCODE) || flags == PRINT_MTIME) {
 				/* set uppercase flag if the first letter of printf-entity is uppercase */
 				modifier_flags |= flags | (format[1] & 0x20 ? 0 : PRINT_FLAG_UPPERCASE);
 				format = p;
@@ -307,14 +381,16 @@ print_item* parse_percent_item(const char** str)
 		/* look for a known token */
 		if (upper && (p = strchr(short_hash, upper))) {
 			assert( (p - short_hash) < (int)(sizeof(hash_ids) / sizeof(unsigned)) );
-			hash_id = hash_ids[p - short_hash];
+			hash_mask = hash_id_to_bit64(hash_ids[p - short_hash]);
 			modifier_flags |= (*format & 0x20 ? 0 : PRINT_FLAG_UPPERCASE);
 		}
 		else if ((p = strchr(short_other, *format))) {
+			static const uint64_t ed2k_aich_hash_mask =
+				hash_id_to_bit64(RHASH_ED2K) | hash_id_to_bit64(RHASH_AICH);
 			assert( (p - short_other) < (int)(sizeof(other_flags) / sizeof(unsigned)) );
 			modifier_flags |= other_flags[p - short_other];
 			if ((modifier_flags & ~PRINT_FLAGS_ALL) == PRINT_ED2K_LINK)
-				hash_id = RHASH_ED2K | RHASH_AICH;
+				hash_mask = ed2k_aich_hash_mask;
 		} else if ((modifier_flags & PRINT_FLAG_URLENCODE) != 0) {
 			/* handle legacy token: '%u' -> '%uf' */
 			modifier_flags |= PRINT_BASENAME;
@@ -323,7 +399,7 @@ print_item* parse_percent_item(const char** str)
 			return 0; /* no valid token found */
 		}
 	}
-	item = new_print_item(modifier_flags, hash_id, NULL);
+	item = new_print_item(modifier_flags, hash_mask, NULL);
 	item->width = width;
 	*str = ++format;
 	return item;
@@ -344,7 +420,7 @@ static int fprint_ed2k_url(FILE* out, struct file_info* info, int print_type)
 	int upper_case = (print_type & PRINT_FLAG_UPPERCASE ? RHPR_UPPERCASE : 0);
 	char buf[104];
 	char* dst = buf;
-	assert(info->sums_flags & RHASH_ED2K);
+	assert(info->hash_mask & hash_id_to_bit64(RHASH_ED2K));
 	assert(info->rctx);
 	if (rsh_fprintf(out, "ed2k://|file|") < 0 || fprint_urlencoded(out, filename, upper_case) < 0)
 		return -1;
@@ -354,7 +430,7 @@ static int fprint_ed2k_url(FILE* out, struct file_info* info, int print_type)
 	*dst++ = '|';
 	rhash_print(dst, info->rctx, RHASH_ED2K, upper_case);
 	dst += 32;
-	if ((info->sums_flags & RHASH_AICH) != 0) {
+	if ((info->hash_mask & hash_id_to_bit64(RHASH_AICH)) != 0) {
 		strcpy(dst, "|h=");
 		rhash_print(dst += 3, info->rctx, RHASH_AICH, RHPR_BASE32 | upper_case);
 		dst += 32;
@@ -432,6 +508,8 @@ int print_line(FILE* out, unsigned out_mode, print_item* list, struct file_info*
 	char buffer[130];
 	int res = 0;
 	unsigned out_flags = (out_mode & FileContentIsUtf8 ? OutForceUtf8 : 0);
+	unsigned line_escaping = 0;
+	unsigned file_escaping = 0;
 #ifdef _WIN32
 	/* switch to binary mode to correctly output binary message digests */
 	int out_fd = _fileno(out);
@@ -444,7 +522,7 @@ int print_line(FILE* out, unsigned out_mode, print_item* list, struct file_info*
 
 		/* output a hash function digest */
 		if (!print_type) {
-			unsigned hash_id = list->hash_id;
+			unsigned hash_id = bit64_to_hash_id(list->hash_mask);
 			int print_flags = (list->flags & PRINT_FLAG_UPPERCASE ? RHPR_UPPERCASE : 0)
 				| (list->flags & PRINT_FLAG_RAW ? RHPR_RAW : 0)
 				| (list->flags & PRINT_FLAG_BASE32 ? RHPR_BASE32 : 0)
@@ -471,20 +549,29 @@ int print_line(FILE* out, unsigned out_mode, print_item* list, struct file_info*
 			case PRINT_ZERO: /* the '\0' character */
 				res = PRINTF_RES(rsh_fprintf(out, "%c", 0));
 				break;
-#ifdef _WIN32
+			case PRINT_ESCAPE_LINE:
+				if (!file_escaping)
+					file_escaping = get_file_escaping_flags(info->file, list->flags, out_flags);
+				if ((file_escaping & list->flags) != 0) {
+					res = PRINTF_RES(rsh_fprintf(out, "\\"));
+					line_escaping = OutEscape; /* start escaping */
+				}
+				break;
 			case PRINT_NEWLINE:
-				res = PRINTF_RES(rsh_fprintf(out, "%s", "\r\n"));
+				res = PRINTF_RES(rsh_fprintf(out, NEWLINE_STR));
+				line_escaping = 0; /* end escaping */
 				break;
-#endif
 			case PRINT_FILEPATH:
-			case PRINT_BASENAME: /* the filename without directory */
-				if ((list->flags & PRINT_FLAG_URLENCODE) != 0) {
-					unsigned fflags = FPathUtf8 | FPathNotNull | (print_type == PRINT_BASENAME ? FPathBaseName : 0);
-					fprint_urlencoded(out, file_get_print_path(info->file, fflags),
-						(list->flags & PRINT_FLAG_UPPERCASE));
-				} else {
-					unsigned pflags = (print_type == PRINT_BASENAME ? OutBaseName : 0);
-					res = PRINTF_RES(fprintf_file_t(out, NULL, info->file, pflags | out_flags));
+				{
+					const unsigned pflags = (list->flags & PRINT_FLAGS_PATH_PARTS) >> 16;
+					assert((PRINT_FLAG_BASENAME >> 16) == FPathBaseName);
+					assert((PRINT_FLAG_DIRNAME >> 16) == FPathDirName);
+					if ((list->flags & PRINT_FLAG_URLENCODE) != 0) {
+						fprint_urlencoded(out, file_get_print_path(info->file, pflags | FPathNotNull | FPathUtf8),
+							(list->flags & PRINT_FLAG_UPPERCASE));
+					} else {
+						res = PRINTF_RES(fprintf_file_t(out, NULL, info->file, pflags | line_escaping | out_flags));
+					}
 				}
 				break;
 			case PRINT_MTIME: /* the last-modified tine of the filename */
@@ -546,7 +633,7 @@ static const char* get_librhash_version(void)
 			v = 0x01030800;
 		else if (algorithm_count == 27)
 			v = 0x01030700;
-		else if (rhash_get_openssl_supported_mask() != RHASH_ERROR)
+		else if (rhash_transmit(12, NULL, 0, 0) != RHASH_ERROR)
 			v = 0x01030600;
 		else if (algorithm_count == 26)
 			return "1.3.[0-5]";
@@ -563,35 +650,36 @@ static const char* get_librhash_version(void)
  */
 void init_hash_info_table(void)
 {
-	unsigned bit;
-	const unsigned fullmask = RHASH_ALL_HASHES | OPT_ED2K_LINK;
-	const unsigned custom_bsd_name = RHASH_RIPEMD160 | RHASH_BLAKE2S | RHASH_BLAKE2B |
+	const uint64_t custom_bsd_name = RHASH_RIPEMD160 | RHASH_BLAKE2S | RHASH_BLAKE2B |
 		RHASH_SHA224 | RHASH_SHA256 | RHASH_SHA384 | RHASH_SHA512;
-	const unsigned short_opt_mask = RHASH_CRC32 | RHASH_MD5 | RHASH_SHA1 | RHASH_TTH | RHASH_ED2K |
-		RHASH_AICH | RHASH_WHIRLPOOL | RHASH_RIPEMD160 | RHASH_GOST12_256 | OPT_ED2K_LINK;
-	const char* short_opt = "cmhteawrgl";
+	const uint64_t short_opt_mask = RHASH_CRC32 | RHASH_MD5 | RHASH_SHA1 | RHASH_TTH | RHASH_ED2K |
+		RHASH_AICH | RHASH_WHIRLPOOL | RHASH_RIPEMD160 | RHASH_GOST12_256;
+	const char* short_opt = "cmhteawrg";
 	print_hash_info* info = hash_info_table;
+	uint64_t hash_mask;
 
 	/* prevent crash on incompatible librhash */
-	if (rhash_count() < RHASH_HASH_COUNT) {
-		rsh_fprintf(stderr, "fatal error: incompatible librhash version is loaded: %s\n", get_librhash_version());
-		rsh_exit(2);
-	} else if (RHASH_HASH_COUNT != rhash_count())
+	RSH_REQUIRE(rhash_count() >= RHASH_HASH_COUNT,
+		"incompatible librhash version is loaded: %s\n", get_librhash_version());
+	if (RHASH_HASH_COUNT != rhash_count())
 		log_warning("inconsistent librhash version is loaded: %s\n", get_librhash_version());
 
 	memset(hash_info_table, 0, sizeof(hash_info_table));
-	for (bit = 1; bit && bit <= fullmask; bit = bit << 1) {
+
+	for (hash_mask = get_all_supported_hash_mask(); hash_mask; hash_mask &= hash_mask - 1) {
+		uint64_t bit64 = hash_mask & -hash_mask;
+		unsigned hash_id = bit64_to_hash_id(bit64);
 		const char* p;
 		char* e;
 		char* d;
 
-		if (!(bit & fullmask))
-			continue;
+		RSH_REQUIRE((info - hash_info_table) < RHASH_HASH_COUNT, "too many hash ids\n");
 
-		info->short_char = ((bit & short_opt_mask) != 0 && *short_opt ?
+		info->hash_id = hash_id;
+		info->short_char = ((bit64 & short_opt_mask) != 0 && *short_opt ?
 			*(short_opt++) : 0);
 
-		info->name = (bit & RHASH_ALL_HASHES ? rhash_get_name(bit) : "ED2K-LINK");
+		info->name = rhash_get_name(hash_id);
 		assert(strlen(info->name) < 19);
 		p = info->name;
 		d = info->short_name;
@@ -612,8 +700,8 @@ void init_hash_info_table(void)
 			}
 		}
 		*d = 0;
-		if ((bit & custom_bsd_name) != 0) {
-			switch (bit) {
+		if ((bit64 & custom_bsd_name) != 0) {
+			switch (hash_id) {
 				case RHASH_RIPEMD160:
 					info->bsd_name = "RMD160";
 					break;
@@ -640,6 +728,9 @@ void init_hash_info_table(void)
 			info->bsd_name = info->name;
 		++info;
 	}
+	assert((info - hash_info_table) == RHASH_HASH_COUNT);
+	assert(info->hash_id == 0);
+	assert(info->name == NULL);
 }
 
 /**
@@ -654,9 +745,9 @@ strbuf_t* init_printf_format(void)
 	strbuf_t* out;
 	const char* fmt;
 	const char* tail = 0;
-	unsigned bit, index = 0;
+	print_hash_info* info;
 	int uppercase;
-	unsigned need_modifier = 0;
+	uint64_t need_modifier = 0;
 	char up_flag;
 	char fmt_modifier = 'b';
 
@@ -667,44 +758,44 @@ strbuf_t* init_printf_format(void)
 	out = rsh_str_new();
 	rsh_str_ensure_size(out, 1024); /* allocate big enough buffer */
 
-	if ((opt.sum_flags & OPT_ED2K_LINK) != 0) {
-		rsh_str_append_n(out, "%l\\n", 4);
-		out->str[1] &= up_flag;
-		return out;
-	}
-	if (opt.sum_flags == 0)
+	if (opt.hash_mask == 0 && opt.fmt != FMT_ED2K_LINK)
 		return out;
 
 	if (opt.fmt == FMT_BSD) {
-		fmt = "\003(%p) = \001\\n";
+		fmt = "\\^\003(%p) = \001\\n";
+	} else if (opt.fmt == FMT_ONE_HASH) {
+		fmt = "\001\\n";
+	} else if (opt.fmt == FMT_ED2K_LINK) {
+		rsh_str_append_n(out, "%l\\n", 4);
+		out->str[1] &= up_flag;
+		return out;
 	} else if (opt.fmt == FMT_MAGNET) {
 		rsh_str_append(out, "magnet:?xl=%s&dn=");
 		rsh_str_append(out, (uppercase ? "%Uf" : "%uf"));
 		fmt = "&xt=urn:\002:\001";
-		need_modifier = RHASH_SHA1;
+		need_modifier = hash_id_to_bit64(RHASH_SHA1);
 		tail = "\\n";
-	} else if (!rhash_data.is_sfv && 0 == (opt.sum_flags & (opt.sum_flags - 1))) {
-		fmt = "\001  %p\\n";
+	} else if (!rhash_data.is_sfv && 0 == (opt.hash_mask & (opt.hash_mask - 1))) {
+		fmt = "\\^\001  %p\\n";
 	} else {
-		rsh_str_append_n(out, "%p", 2);
-		fmt = (rhash_data.is_sfv ? " \001" : "  \001");
+		rsh_str_append_n(out, "\\^%p", 4);
+		fmt = " \001";
 		tail = "\\n";
 	}
 	if ((opt.flags & OPT_FMT_MODIFIERS) != 0)
 	{
-		need_modifier = 0xffffffff;
+		need_modifier = (uint64_t)-1;
 		fmt_modifier = (opt.flags & OPT_HEX ? 'x' : opt.flags & OPT_BASE32 ? 'b' : 'B');
 	}
 
 	/* loop by message digests */
-	for (bit = 1 << index; bit && bit <= opt.sum_flags; bit = bit << 1, index++) {
+	for (info = hash_info_table; info->hash_id; info++) {
 		const char* p;
-		print_hash_info* info;
+		uint64_t bit64 = hash_id_to_bit64(info->hash_id);
 
-		if ((bit & opt.sum_flags) == 0)
+		if ((bit64 & opt.hash_mask) == 0)
 			continue;
 		p = fmt;
-		info = &hash_info_table[index];
 
 		/* ensure the output buffer have enough space */
 		rsh_str_ensure_size(out, out->len + 256);
@@ -718,7 +809,7 @@ strbuf_t* init_printf_format(void)
 			switch ((int)*(p++)) {
 				case 1:
 					out->str[out->len++] = '%';
-					if ( (bit & need_modifier) != 0 )
+					if ( (bit64 & need_modifier) != 0 )
 						out->str[out->len++] = fmt_modifier;
 					if (info->short_char)
 						out->str[out->len++] = info->short_char & up_flag;
@@ -732,7 +823,7 @@ strbuf_t* init_printf_format(void)
 					}
 					break;
 				case 2:
-					rsh_str_append(out, rhash_get_magnet_name(bit));
+					rsh_str_append(out, rhash_get_magnet_name(info->hash_id));
 					break;
 				case 3:
 					rsh_str_append(out, info->bsd_name);
@@ -765,8 +856,8 @@ strbuf_t* init_printf_format(void)
 int print_sfv_header_line(FILE* out, unsigned out_mode, file_t* file)
 {
 	char buf[24];
-	unsigned out_flags = (out_mode & FileContentIsUtf8 ? OutForceUtf8 : 0);
-
+	unsigned out_flags = (out_mode & FileContentIsUtf8 ?
+		OutForceUtf8 | OutEscapePrefixed : OutEscapePrefixed);
 	/* skip stdin stream and message-texts passed by command-line */
 	if (FILE_ISSPECIAL(file))
 		return 0;
