@@ -3,6 +3,7 @@
 #define HASH_PRINT_H
 
 #include <stdio.h>
+#include <stdint.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -15,6 +16,7 @@ typedef struct print_hash_info
 {
 	const char* name;
 	const char* bsd_name;
+	unsigned hash_id;
 	char short_name[20];
 	char short_char;
 } print_hash_info;
@@ -31,7 +33,7 @@ void init_hash_info_table(void);
 struct strbuf_t* init_printf_format(void);
 
 /* formatted output of message digests and file information */
-struct print_item* parse_print_string(const char* format, unsigned* sum_mask);
+struct print_item* parse_print_string(const char* format, uint64_t* hash_mask);
 int print_line(FILE* out, unsigned out_mode, struct print_item* list, struct file_info* info);
 void free_print_list(struct print_item* list);
 
