@@ -46,6 +46,86 @@ static int count_printed_size(const char* format, const char* path, unsigned out
 	return format_length + (IS_UTF8() || (output_flags & OutForceUtf8) ? count_utf8_symbols(path) : strlen(path));
 }
 
+/**
+ * Test if string has CR ('\n') or LF ('\r') character.
+ *
+ * @param str the string to test
+ * @return non-zero if str has CR or LF
+ */
+static int has_cr_or_lf(const char* str)
+{
+	size_t i;
+	for (i = 0; (((int)str[i]) * ((int)str[i] - '\r') * ((int)str[i] - '\n')) != 0; i++);
+	return (str[i] != 0);
+}
+
+/**
+ * Print string with escaped characters to the specified stream.
+ * Only characters '\r', '\n' and '\\' are escaped by prefixing
+ * the escape symbol '\\'.
+ *
+ * @param out the stream to print to
+ * @param str the string to print with escaping
+ * @param output_flags bitmask containing OutForceUtf8 and OutCountSymbols flags
+ * @return the number of characters printed, -1 on fail with error code stored in errno
+ */
+static int fprint_escaped(FILE* out, const char* str, unsigned output_flags)
+{
+	char buffer[1024];
+	const size_t write_max = sizeof(buffer) - 8;
+	const char* in = str;
+	size_t escaped = 0;
+	size_t pos = 0;
+	if ((output_flags & OutEscapePrefixed) != 0) {
+		if (has_cr_or_lf(str)) {
+			escaped++;
+			buffer[pos++] = '\\';
+		} else if (rsh_fprintf(out, "%s", str) < 0)
+			return -1;
+		else
+			in = "";
+	}
+	while (*in)
+	{
+		int special_char = 0;
+		switch (*in)
+		{
+			case '\\':
+				special_char = '\\';
+				break;
+			case '\n':
+				special_char = 'n';
+				break;
+			case '\r':
+				special_char = 'r';
+				break;
+		}
+		if (special_char) {
+			escaped++;
+			buffer[pos++] = '\\';
+			buffer[pos++] = (char)special_char;
+		} else
+			buffer[pos++] = *in;
+		in++;
+		if (*in)
+		{
+			if (pos < write_max)
+				continue;
+			/* copy a possible uft8 symbol till its end */
+			if (((*in & 0xc0) == 0x80) && pos < (write_max + 4))
+				continue;
+		}
+		assert(pos < sizeof(buffer));
+		buffer[pos] = '\0';
+		if (rsh_fprintf(out, "%s", buffer) < 0)
+			return -1;
+		pos = 0;
+	}
+	if ((output_flags & OutCountSymbols) != 0)
+		return escaped + count_printed_size(NULL, str, output_flags);
+	return 0;
+}
+
 /**
  * Print formatted file path to the specified stream.
  *
@@ -57,27 +137,30 @@ static int count_printed_size(const char* format, const char* path, unsigned out
  */
 int fprintf_file_t(FILE* out, const char* format, struct file_t* file, unsigned output_flags)
 {
-	unsigned basename_bit = output_flags & FPathBaseName;
+	unsigned name_bits = output_flags & (FPathBaseName | FPathDirName);
 #ifdef _WIN32
 	const char* print_path;
 	if (!file->real_path) {
-		print_path = file_get_print_path(file, FPathPrimaryEncoding | FPathNotNull | basename_bit);
+		print_path = file_get_print_path(file, FPathPrimaryEncoding | FPathNotNull | name_bits);
 	} else {
 		unsigned ppf = ((output_flags & OutForceUtf8) || (opt.flags & OPT_UTF8) ? FPathUtf8 | FPathNotNull : FPathPrimaryEncoding);
 		assert(file->real_path != NULL);
 		assert((int)OutBaseName == (int)FPathBaseName);
-		print_path = file_get_print_path(file, ppf | basename_bit);
+		print_path = file_get_print_path(file, ppf | name_bits);
 		if (!print_path) {
-			print_path = file_get_print_path(file, FPathUtf8 | FPathNotNull | basename_bit);
+			print_path = file_get_print_path(file, FPathUtf8 | FPathNotNull | name_bits);
 			assert(print_path);
 			assert(!(opt.flags & OPT_UTF8));
 		}
 	}
 #else
-	const char* print_path = file_get_print_path(file, FPathPrimaryEncoding | FPathNotNull | basename_bit);
+	const char* print_path = file_get_print_path(file, FPathPrimaryEncoding | FPathNotNull | name_bits);
 	assert((int)OutBaseName == (int)FPathBaseName);
+	assert((int)OutDirName == (int)FPathDirName);
 	assert(print_path);
 #endif
+	if (!format && (output_flags & (OutEscape | OutEscapePrefixed)) != 0)
+		return fprint_escaped(out, print_path, output_flags);
 	if (rsh_fprintf(out, (format ? format : "%s"), print_path) < 0)
 		return -1;
 	if ((output_flags & OutCountSymbols) != 0)
@@ -171,10 +254,43 @@ void log_msg(const char* format, ...)
  */
 void log_msg_file_t(const char* format, struct file_t* file)
 {
-	fprintf_file_t(rhash_data.log, format, file, OutDefaultFlags);
+	fprintf_file_t(rhash_data.log, format, file, OutEscapePrefixed);
 	fflush(rhash_data.log);
 }
 
+/**
+ * Print a formatted error message and exit.
+ *
+ * @param format the format string
+ */
+void die(const char* format, ...)
+{
+	va_list ap;
+	va_start(ap, format);
+	print_log_prefix();
+	log_va_msg(format, ap);
+	va_end(ap);
+	rsh_exit(2);
+}
+
+/**
+ * Print fatal error message with source_file:line: prefix
+ * and exit.
+ *
+ * @param file source file name
+ * @param line source code line
+ * @param format the format string
+ */
+void fatal_error_impl(const char* file, int line, const char* format, ...)
+{
+	va_list ap;
+	va_start(ap, format);
+	rsh_fprintf(rhash_data.log, "%s: %s:%d: ", PROGRAM_NAME, file, line);
+	log_va_msg(format, ap);
+	va_end(ap);
+	rsh_exit(2);
+}
+
 /**
  * Print a formatted error message to the program log.
  *
@@ -198,7 +314,7 @@ void log_error_file_t(struct file_t* file)
 {
 	int file_errno = errno;
 	print_log_prefix();
-	fprintf_file_t(rhash_data.log, "%s", file, OutDefaultFlags);
+	fprintf_file_t(rhash_data.log, "%s", file, OutEscapePrefixed);
 	rsh_fprintf(rhash_data.log, ": %s\n", strerror(file_errno));
 	fflush(rhash_data.log);
 }
@@ -211,7 +327,7 @@ void log_error_file_t(struct file_t* file)
 void log_error_msg_file_t(const char* format, struct file_t* file)
 {
 	print_log_prefix();
-	fprintf_file_t(rhash_data.log, format, file, OutDefaultFlags);
+	fprintf_file_t(rhash_data.log, format, file, OutEscapePrefixed);
 	fflush(rhash_data.log);
 }
 
@@ -240,9 +356,6 @@ struct percents_t
 };
 static struct percents_t percents;
 
-/* the hash functions, which needs to be reported first on mismatch */
-#define REPORT_FIRST_MASK (RHASH_MD5 | RHASH_SHA256 | RHASH_SHA512)
-
 /**
  * Print verbose error on a message digest mismatch.
  *
@@ -272,12 +385,17 @@ static int print_verbose_hash_check_error(struct file_info* info)
 	}
 
 	if (HpWrongHashes & info->hp->bit_flags) {
-		unsigned reported = 0;
+		/* bits-mask for hash functions to be reported first on mismatch */
+		static const uint64_t report_first_mask =
+			hash_id_to_bit64(RHASH_MD5) |
+			hash_id_to_bit64(RHASH_SHA256) |
+			hash_id_to_bit64(RHASH_SHA512);
+		uint64_t reported = 0;
 		int i;
 		for (i = 0; i < info->hp->hashes_num; i++) {
 			struct hash_value* hv = &info->hp->hashes[i];
 			char* expected_hash = info->hp->line_begin + hv->offset;
-			unsigned hid = hv->hash_id;
+			uint64_t hid = hv->hash_mask;
 			int pflags;
 			if ((info->hp->wrong_hashes & (1 << i)) == 0)
 				continue;
@@ -287,10 +405,10 @@ static int print_verbose_hash_check_error(struct file_info* info)
 			/* if can't detect precise hash */
 			if ((hid & (hid - 1)) != 0) {
 				/* guess the hash id */
-				if (hid & opt.sum_flags) hid &= opt.sum_flags;
+				if (hid & opt.hash_mask) hid &= opt.hash_mask;
 				if (hid & ~info->hp->found_hash_ids) hid &= ~info->hp->found_hash_ids;
 				if (hid & ~reported) hid &= ~reported; /* avoid repeating */
-				if (hid & REPORT_FIRST_MASK) hid &= REPORT_FIRST_MASK;
+				if (hid & report_first_mask) hid &= report_first_mask;
 				hid &= -(int)hid; /* take the lowest bit */
 			}
 			assert(hid != 0 && (hid & (hid - 1)) == 0); /* single bit only */
@@ -317,7 +435,7 @@ static int print_verbose_hash_check_error(struct file_info* info)
  */
 static int print_aligned_filepath(FILE* out, struct file_info* info)
 {
-	int symbols_count = fprintf_file_t(out, NULL, info->file, OutCountSymbols);
+	int symbols_count = fprintf_file_t(out, NULL, info->file, OutCountSymbols | OutEscapePrefixed);
 	if (symbols_count >= 0) {
 		char buf[56];
 		int spaces_count = (symbols_count <= 51 ? 52 - symbols_count : 1);
@@ -345,14 +463,14 @@ static int print_check_result(struct file_info* info, int print_name, int print_
 		if (info->processing_result < 0) {
 			/* print error to stdout */
 			res = PRINTF_RES(rsh_fprintf(rhash_data.out, "%s\n", strerror(saved_errno)));
-		} else if (!HP_FAILED(info->hp->bit_flags) || !(opt.flags & OPT_VERBOSE)) {
+		} else if (HP_FAILED(info->hp->bit_flags) && opt.verbose) {
+			res = print_verbose_hash_check_error(info);
+		} else {
 			res = PRINTF_RES(rsh_fprintf(rhash_data.out, (!HP_FAILED(info->hp->bit_flags) ?
 				/* TRANSLATORS: printed when all message digests match, use at least 3 characters to overwrite "99%" */
 				_("OK \n") :
 				/* TRANSLATORS: ERR (short for 'error') is printed on a message digest mismatch */
 				_("ERR\n"))));
-		} else {
-			res = print_verbose_hash_check_error(info);
 		}
 	}
 	if (fflush(rhash_data.out) < 0)
@@ -638,7 +756,7 @@ int print_verifying_header(file_t* file)
 	{
 		char dash_line[84];
 		/* TRANSLATORS: the line printed before a hash file is verified */
-		int count = fprintf_file_t(rhash_data.out, _("\n--( Verifying %s )"), file, OutCountSymbols);
+		int count = fprintf_file_t(rhash_data.out, _("\n--( Verifying %s )"), file, OutCountSymbols | OutEscapePrefixed);
 		int tail_dash_len = (0 < count && count < 81 ? 81 - count : 2);
 		int res = rsh_fprintf(rhash_data.out, "%s\n", str_set(dash_line, '-', tail_dash_len));
 		if (res >= 0)
@@ -681,6 +799,32 @@ int print_check_stats(void)
 	return (fflush(rhash_data.out) < 0 ? -1 : res);
 }
 
+/**
+ * Print used algorithms, but only on very verbose level
+ *
+ * @param out the stream to print the result to
+ * @param hash_mask bit mask for calculated message digests to print
+ * @return 0 on success, -1 on fail with error code stored in errno
+ */
+int print_verbose_algorithms(FILE* out, uint64_t hash_mask)
+{
+	const char* prefix;
+	if (!hash_mask || opt.verbose <= 1)
+		return 0;
+	prefix = _("Calculating: ");
+	while (hash_mask) {
+		uint64_t bit64 = hash_mask & -hash_mask;
+		const char* hash_name = rhash_get_name(bit64_to_hash_id(bit64));
+		if (rsh_fprintf(out, "%s%s", prefix, hash_name) < 0)
+			return -1;
+		prefix = ", ";
+		hash_mask ^= bit64;
+	}
+	if (rsh_fprintf(out, "\n") < 0)
+		return -1;
+	return (fflush(out) < 0 ? -1 : 0);
+}
+
 /**
  * Print file processing times.
  */
